<?php

namespace Drupal\commerce_affirm\Plugin\Commerce\PaymentGateway;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_affirm\Event\AffirmEvents;
use Drupal\commerce_affirm\Event\AffirmLocaleEvent;
use Drupal\commerce_affirm\Event\AffirmTransactionDataPreSend;
use Drupal\commerce_checkout\Event\CheckoutEvents;
use Drupal\commerce_order\Adjustment;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Event\OrderEvent;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsVoidsInterface;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\commerce_promotion\Entity\Promotion;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the Affirm Redirect payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "affirm_redirect",
 *   label = @Translation("Affirm (modal or redirect)"),
 *   display_label = @Translation("Affirm"),
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_affirm\PluginForm\RedirectForm",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class Redirect extends OffsitePaymentGatewayBase implements SupportsRefundsInterface, SupportsVoidsInterface, SupportsAuthorizationsInterface {

  /**
   * The logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The module extension list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->logger = $container->get('logger.channel.commerce_affirm');
    $instance->httpClient = $container->get('http_client');
    $instance->eventDispatcher = $container->get('event_dispatcher');
    $instance->moduleExtensionList = $container->get('extension.list.module');
    $instance->moduleHandler = $container->get('module_handler');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    $return = [
      'window_mode' => 'modal',
      'public_key' => '',
      'private_key' => '',
      'default_locale' => 'en_US',
      'country_code' => 'USA',
      'prepend_display_name_with_logo' => 'none',
      'hide_display_name_use_logo' => FALSE,
      'log' => FALSE,
    ];
    return $return + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['window_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Window mode'),
      '#options' => [
        'modal' => $this->t('Modal (on checkout review)'),
        'redirect' => $this->t('Redirect'),
      ],
      '#default_value' => $this->configuration['window_mode'],
    ];
    // Public key.
    $form['public_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Public Key'),
      '#description' => $this->t('Add your public key for the gateway.'),
      '#default_value' => $this->configuration['public_key'],
      '#required' => TRUE,
    ];
    // Private key.
    $form['private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private Key'),
      '#description' => $this->t('Add your private key for the gateway.'),
      '#default_value' => $this->configuration['private_key'],
      '#required' => TRUE,
    ];
    // Default locale.
    $form['default_locale'] = [
      '#type' => 'select',
      '#options' => [
        'en_US' => $this->t('USA - English'),
        'en_CA' => $this->t('Canada - English'),
        'fr_CA' => $this->t('Canada - French'),
      ],
      '#title' => $this->t('Default locale'),
      '#default_value' => $this->configuration['default_locale'],
    ];
    // Country code.
    $form['country_code'] = [
      '#type' => 'select',
      '#options' => [
        'USA' => $this->t('USA'),
        'CAN' => $this->t('Canada'),
      ],
      '#title' => $this->t('Country code'),
      '#default_value' => $this->configuration['country_code'],
    ];
    $form['prepend_display_name_with_logo'] = [
      '#type' => 'select',
      '#title' => $this->t('Prepend the display name with a logo'),
      '#description' => $this->t('Preview the logos in the <a href=":url" target="_blank">Logos</a> homepage.', [
        ':url' => 'https://businesshub.affirm.com/hc/en-us/articles/4425266541204-Logos-Banners-and-Buttons#h_01HA4YQMWW2FZKAT9P8KMFJJT7',
      ]),
      '#options' => [
        'none' => $this->t('None'),
        'black_logo_white_bg' => $this->t('Black-Indigo Logo'),
        'white_logo_solid_bg' => $this->t('White Logo'),
        'all_black_logo_white_bg' => $this->t('All Black Logo'),
        'blue_solid_circle_white_bg' => $this->t('Indigo Solid Circle'),
        'black_solid_circle_white_bg' => $this->t('Black Solid Circle'),
      ],
      '#default_value' => $this->configuration['prepend_display_name_with_logo'] ?? 'none',
    ];
    $form['hide_display_name_use_logo'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide the display name and just use the logo.'),
      '#default_value' => $this->configuration['hide_display_name_use_logo'] ?? FALSE,
      '#states' => [
        'invisible' => [
          ':input[name="configuration[' . $this->pluginId . '][prepend_display_name_with_logo]"]' => ['value' => 'none'],
        ],
      ],
    ];
    $form['log'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Log API and related data for debugging'),
      '#default_value' => $this->configuration['log'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['window_mode'] = $values['window_mode'];
      $this->configuration['public_key'] = $values['public_key'];
      $this->configuration['private_key'] = $values['private_key'];
      $this->configuration['default_locale'] = $values['default_locale'];
      $this->configuration['country_code'] = $values['country_code'];
      $this->configuration['prepend_display_name_with_logo'] = $values['prepend_display_name_with_logo'];
      $this->configuration['hide_display_name_use_logo'] = $values['hide_display_name_use_logo'];
      $this->configuration['log'] = $values['log'];
    }
  }

  /**
   * Place the order.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @return string
   *   The next step id.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function placeOrder(OrderInterface $order): string {
    $step_id = 'complete';
    $order->set('checkout_step', $step_id);
    // Notify other modules.
    $event = new OrderEvent($order);
    $this->eventDispatcher->dispatch($event, CheckoutEvents::COMPLETION);
    $order->getState()->applyTransitionById('place');
    $order->save();
    return $step_id;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request): void {
    $data = [
      'transaction_id' => $request->request->get('checkout_token'),
      'order_id' => $order->id(),
    ];
    try {
      if ($this->configuration['log']) {
        $this->logger->debug('Authorization request for order ID @id: <pre>@data</pre>', [
          '@data' => print_r($data, TRUE),
          '@id' => $order->id(),
        ]);
      }

      $response = $this->apiRequest('authorization', $data);
    }
    catch (\Exception $e) {
      $this->logger->warning($e->getMessage());
      throw new PaymentGatewayException($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
    }

    if (empty($response)) {
      $this->messenger()->addError($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
      throw new PaymentGatewayException($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
    }

    if ($response->getStatusCode() == 200) {
      $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
      $response_json = json_decode($response->getBody());

      if ($this->configuration['log']) {
        $this->logger->debug('Authorization response for order ID @id: <pre>@data</pre>', [
          '@data' => print_r($response_json, TRUE),
          '@id' => $order->id(),
        ]);
      }

      // Checks if the response contains an error.
      if (isset($response_json->status_code)) {
        // Display an error message and remain on the same page.
        $this->messenger()->addError($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
        throw new PaymentGatewayException($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
      }

      /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
      $payment = $payment_storage->create([
        'state' => 'authorization',
        'amount' => $order->getTotalPrice(),
        'payment_gateway' => $this->parentEntity->id(),
        'order_id' => $order->id(),
        'remote_id' => $response_json->id,
        'remote_status' => 'authorize',
      ]);
      $payment->save();

      // Capture the payment if we can determine that the checkout flow says it
      // should be captured via the payment process pane.
      /** @var \Drupal\commerce_checkout\Entity\CheckoutFlowInterface $checkout_flow */
      if ($checkout_flow = $order->get('checkout_flow')->entity) {
        $configuration = $checkout_flow->getPlugin()->getConfiguration();
        if (!empty($configuration['panes']['payment_process']['capture'])) {
          $this->capturePayment($payment);
        }
      }

      // Only place the order when using the modal; redirected customers will
      // process normally via the checkout flow.
      if ($this->getWindowMode() == 'modal') {
        $step_id = $this->placeOrder($order);
        throw new NeedsRedirectException(Url::fromRoute('commerce_checkout.form', [
          'commerce_order' => $order->id(),
          'step' => $step_id,
        ])->toString());
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, ?Price $amount = NULL): void {
    if (!empty($amount) && $amount->lessThan($payment->getAmount())) {
      $this->logger->warning($this->t('Partial capture attempted for payment @id.', ['@id' => $payment->id()]));
      throw new PaymentGatewayException($this->t('Affirm does not support partial captures.'));
    }
    $amount = $payment->getAmount();

    $data = [
      'amount' => $this->minorUnitsConverter->toMinorUnits($amount),
    ];
    try {
      if ($this->configuration['log']) {
        $this->logger->debug('Capture request for payment @id: <pre>@data</pre>', [
          '@id' => $payment->getRemoteId(),
          '@data' => print_r($data, TRUE),
        ]);
      }

      $response = $this->apiRequest('capture', $data, $payment);
    }
    catch (\Exception $e) {
      $this->logger->warning($e->getMessage());
      throw new PaymentGatewayException($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $response_json = json_decode($response->getBody());

    if ($this->configuration['log']) {
      $this->logger->debug('Capture response for payment @id: <pre>@data</pre>', [
        '@id' => $payment->getRemoteId(),
        '@data' => print_r($response_json, TRUE),
      ]);
    }

    if (empty($response_json) || isset($response_json->status_code)) {
      throw new PaymentGatewayException($this->t('We could not complete your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $payment->setState('completed');
    $payment->setRemoteState($response_json->type);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment): void {
    try {
      if ($this->configuration['log']) {
        $this->logger->debug('Void request for payment @id: <pre>@data</pre>', [
          '@id' => $payment->getRemoteId(),
          '@data' => print_r([], TRUE),
        ]);
      }

      $response = $this->apiRequest('void', [], $payment);
    }
    catch (\Exception $e) {
      $this->logger->warning($e->getMessage());
      throw new PaymentGatewayException($this->t('We could not void your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $response_json = json_decode($response->getBody());

    if ($this->configuration['log']) {
      $this->logger->debug('Void response for payment @id: <pre>@data</pre>', [
        '@id' => $payment->getRemoteId(),
        '@data' => print_r($response_json, TRUE),
      ]);
    }

    if (empty($response_json) || isset($response_json->status_code)) {
      throw new PaymentGatewayException($this->t('We could not void your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $payment->setState('authorization_voided');
    $payment->setRemoteState($response_json->type);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, ?Price $amount = NULL): void {
    if (!$amount) {
      $amount = $payment->getAmount();
    }
    $data = [
      'amount' => $this->minorUnitsConverter->toMinorUnits($amount),
    ];
    try {
      if ($this->configuration['log']) {
        $this->logger->debug('Refund request for payment @id: <pre>@data</pre>', [
          '@id' => $payment->getRemoteId(),
          '@data' => print_r($data, TRUE),
        ]);
      }

      $response = $this->apiRequest('refund', $data, $payment);
    }
    catch (\Exception $e) {
      $this->logger->warning($e->getMessage());
      throw new PaymentGatewayException($this->t('We could not refund your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $response_json = json_decode($response->getBody());

    if ($this->configuration['log']) {
      $this->logger->debug('Refund response for payment @id: <pre>@data</pre>', [
        '@id' => $payment->getRemoteId(),
        '@data' => print_r($response_json, TRUE),
      ]);
    }

    if (empty($response_json) || isset($response_json->status_code)) {
      throw new PaymentGatewayException($this->t('We could not refund your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }
    $payment->setRemoteState($response_json->type);
    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

  /**
   * Updates the given payment.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param \Drupal\commerce_payment\Entity\PaymentInterface $payment
   *   The payment to refund.
   *
   * @throws \Drupal\commerce_payment\Exception\PaymentGatewayException
   *   Thrown when the transaction fails for any reason.
   */
  public function updatePayment(OrderInterface $order, PaymentInterface $payment): void {
    $data = [
      'order_id' => $order->getOrderNumber(),
    ];
    try {
      if ($this->configuration['log']) {
        $this->logger->debug('Update request for payment @id: <pre>@data</pre>', [
          '@id' => $payment->getRemoteId(),
          '@data' => print_r($data, TRUE),
        ]);
      }

      $response = $this->apiRequest('update', $data, $payment);
    }
    catch (\Exception $e) {
      $this->logger->warning($e->getMessage());
      throw new PaymentGatewayException($this->t('We could not update your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
    $response_json = json_decode($response->getBody());

    if ($this->configuration['log']) {
      $this->logger->debug('Update response for payment @id: <pre>@data</pre>', [
        '@id' => $payment->getRemoteId(),
        '@data' => print_r($response_json, TRUE),
      ]);
    }

    if (empty($response_json) || isset($response_json->status_code)) {
      throw new PaymentGatewayException($this->t('We could not update your payment with Affirm. Please try again or contact us if the problem persists.'));
    }
  }

  /**
   * Returns the base URL to Affirm API server.
   *
   * @return string
   *   The Base URL to use to submit requests to the Affirm API server.
   */
  public function serverUrl(): string {
    if ($this->getMode() === 'test') {
      return 'https://api.global-sandbox.affirm.com/api/v1/transactions';
    }
    else {
      return 'https://api.global.affirm.com/api/v1/transactions';
    }
  }

  /**
   * Returns the base URL of the Affirm dashboard.
   *
   * @return string
   *   The base URL of the Affirm dashboard.
   */
  public function getDashboardBaseUrl(): string {
    $base_url = $this->getMode() === 'test' ? 'https://sandbox.affirm.com' :
      'https://www.affirm.com';
    if ($this->getCountryCode() === 'CAN') {
      $base_url = $this->getMode() === 'test' ? 'https://sandbox.affirm.ca' :
        'https://www.affirm.ca';
    }

    return $base_url;
  }

  /**
   * Returns the URL to the Affirm server determined by the transaction type.
   *
   * @param string $request_type
   *   The transaction request type (authorization/capture/refund/void).
   * @param \Drupal\commerce_payment\Entity\PaymentInterface|null $payment
   *   The payment, NULL if not set.
   *
   * @return string|bool
   *   The URL to use to submit requests to the Affirm server.
   */
  public function apiServerUrl(string $request_type, ?PaymentInterface $payment = NULL) {
    $api_url = $this->serverUrl();

    switch ($request_type) {
      case 'authorization':
        return $api_url;

      case 'capture':
      case 'refund':
      case 'void':
        if (empty($payment)) {
          $this->messenger()->addStatus($this->t('You can not create a @request_type request without specifying a payment.', [
            '@request_type' => $request_type,
          ]));
          return FALSE;
        }
        $transaction_id = $payment->getRemoteId();
        return "$api_url/$transaction_id/$request_type";

      case 'update':
      case 'read':
        if (empty($payment)) {
          $this->messenger()->addStatus($this->t('You can not create a @request_type request without specifying a payment.', [
            '@request_type' => $request_type,
          ]));
          return FALSE;
        }
        $transaction_id = $payment->getRemoteId();
        return "$api_url/$transaction_id";
    }

    return FALSE;
  }

  /**
   * Perform a request against the Affirm API.
   *
   * @param string $request_type
   *   The transaction request type (authorization/capture/refund/void).
   * @param array $data
   *   Data to send to Affirm.
   * @param \Drupal\commerce_payment\Entity\PaymentInterface|null $payment
   *   The payment, NULL if not set.
   *
   * @return \Psr\Http\Message\ResponseInterface
   *   The Affirm response.
   */
  public function apiRequest(string $request_type, array $data, ?PaymentInterface $payment = NULL) {
    $options = [
      'headers' => [
        'Content-Type' => 'application/json',
        'country-code' => $this->getCountryCode(),
      ],
      'auth' => [
        $this->getPublicKey(),
        $this->getPrivateKey(),
      ],
    ];

    // A json parameter should not be added for a Void request.
    if ($request_type !== 'void') {
      $options['json'] = $data;
    }

    return $this->httpClient->post($this->apiServerUrl($request_type, $payment), $options);
  }

  /**
   * Gets the Affirm public key.
   *
   * @return string
   *   Returns the public key.
   */
  public function getPublicKey(): string {
    return $this->configuration['public_key'];
  }

  /**
   * Gets the Affirm private key.
   *
   * @return string
   *   Returns the private key.
   */
  public function getPrivateKey(): string {
    return $this->configuration['private_key'];
  }

  /**
   * Gets the Affirm default locale.
   *
   * @return string
   *   Returns the default locale.
   */
  public function getDefaultLocale(): string {
    return $this->configuration['default_locale'];
  }

  /**
   * Gets the Affirm default locale altered by other modules.
   *
   * @return string
   *   Returns the default locale altered by other modules.
   */
  public function getLocale(): string {
    $locale = $this->getDefaultLocale();

    // Allow modules to alter the default locale.
    $event = new AffirmLocaleEvent($locale);
    $this->eventDispatcher->dispatch($event, AffirmEvents::AFFIRM_LOCALE);
    $locale = $event->getLocale();

    return $locale;
  }

  /**
   * Gets the Affirm country code.
   *
   * @return string
   *   Returns the country code.
   */
  public function getCountryCode(): string {
    return $this->configuration['country_code'];
  }

  /**
   * Gets the URL of the Affirm script.
   *
   * @return string
   *   The URL of the Affirm script.
   */
  public function getScriptUrl(): string {
    $country_code = $this->getCountryCode();

    if ($this->getMode() === 'test') {
      return $country_code === 'CAN' ?
        'https://cdn1-sandbox.affirm.ca/js/v2/affirm.js' :
        'https://cdn1-sandbox.affirm.com/js/v2/affirm.js';
    }
    else {
      return $country_code === 'CAN' ?
        'https://cdn1.affirm.ca/js/v2/affirm.js' :
        'https://cdn1.affirm.com/js/v2/affirm.js';
    }
  }

  /**
   * Gets the Affirm window mode.
   *
   * @return string
   *   Returns the window mode.
   */
  public function getWindowMode(): string {
    return $this->configuration['window_mode'];
  }

  /**
   * Gets the Affirm log settings.
   *
   * @return string
   *   Returns the og settings.
   */
  public function getLogSettings(): string {
    return $this->configuration['log'];
  }

  /**
   * Gets the Affirm checkout settings.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @return array
   *   The Affirm checkout settings.
   */
  public function getAffirmCheckoutSettings(OrderInterface $order): array {
    $order_total = $order->getTotalPrice();

    $commerce_module_info = $this->moduleExtensionList->getExtensionInfo('commerce');
    $affirm_module_info = $this->moduleExtensionList->getExtensionInfo('commerce_affirm');
    if ($commerce_module_info['version'] == '') {
      $commerce_module_info['version'] = '8.x-1.??';
    }

    // Creates a data array to be passed onto the request.
    $checkout_settings = [
      'merchant' => [
        'user_confirmation_url' => Url::fromRoute('commerce_payment.checkout.return', [
          'commerce_order' => $order->id(),
          'step' => $this->getWindowMode() === 'modal' ? 'review' : 'payment',
        ], ['absolute' => TRUE])->toString(),
        'user_confirmation_url_action' => 'POST',
        'user_cancel_url' => Url::fromRoute('commerce_payment.checkout.cancel', [
          'commerce_order' => $order->id(),
          'step' => $this->getWindowMode() === 'modal' ? 'review' : 'payment',
        ], ['absolute' => TRUE])->toString(),
      ],
      'billing' => [],
      'items' => [],
      'shipping' => [],
      'metadata' => [
        'platform_type' => 'Drupal',
        'platform_version' => 'Drupal ' . \Drupal::VERSION . '; Drupal Commerce ' . $commerce_module_info['version'],
        'platform_affirm' => $affirm_module_info['version'],
        'shipping_type' => '',
        'mode' => $this->getWindowMode(),
      ],
      'shipping_amount' => 0,
      'tax_amount' => 0,
      'total' => $this->minorUnitsConverter->toMinorUnits($order_total),
      'currency' => $order_total->getCurrencyCode(),
    ];

    // Adds information about the billing profile.
    if ($billing_profile = $order->getBillingProfile()) {
      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $billing_address */
      $billing_address = $billing_profile->get('address')->first();
      $checkout_settings['billing'] = [
        'name' => [
          'full' => $billing_address->getGivenName() . ' ' . $billing_address->getFamilyName(),
          'first' => $billing_address->getGivenName(),
          'last' => $billing_address->getFamilyName(),
        ],
        'address' => [
          'line1' => $billing_address->getAddressLine1(),
          'line2' => $billing_address->getAddressLine2(),
          'city' => $billing_address->getLocality(),
          'state' => $billing_address->getAdministrativeArea(),
          'zipcode' => $billing_address->getPostalCode(),
          'country' => $billing_address->getCountryCode(),
        ],
        'email' => $order->getEmail(),
        'phone_number' => '',
      ];
    }

    // Use the shipping profile if the shipping module exists and the order
    // contain a shipping profile otherwise use the Billing profile.
    $order_profiles = $order->collectProfiles();
    $profile = $order_profiles['shipping'] ?? $billing_profile;
    if ($profile) {
      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $shipping_address */
      $shipping_address = $profile->get('address')->first();
      $checkout_settings['shipping'] = [
        'name' => [
          'full' => $shipping_address->getGivenName() . ' ' . $shipping_address->getFamilyName(),
          'first' => $shipping_address->getGivenName(),
          'last' => $shipping_address->getFamilyName(),
        ],
        'address' => [
          'line1' => $shipping_address->getAddressLine1(),
          'line2' => $shipping_address->getAddressLine2(),
          'city' => $shipping_address->getLocality(),
          'state' => $shipping_address->getAdministrativeArea(),
          'zipcode' => $shipping_address->getPostalCode(),
          'country' => $shipping_address->getCountryCode(),
        ],
        'email' => $order->getEmail(),
        'phone_number' => '',
      ];
    }

    foreach ($order->getItems() as $order_item) {
      $checkout_settings['items'][] = [
        'sku' => $order_item->getPurchasedEntity()->getSku(),
        'qty' => $order_item->getQuantity(),
        'display_name' => $order_item->getTitle(),
        'item_url' => $order_item->getPurchasedEntity()->toUrl()->setAbsolute()->toString(),
        'item_image_url' => '',
        'unit_price' => $this->minorUnitsConverter->toMinorUnits($order_item->getUnitPrice()),
      ];
      foreach ($order_item->getAdjustments() as $adjustment) {
        $this->handleAffirmCheckoutAdjustments($order, $adjustment, $checkout_settings);
      }
    }

    foreach ($order->getAdjustments() as $adjustment) {
      $this->handleAffirmCheckoutAdjustments($order, $adjustment, $checkout_settings);
    }
    sort($checkout_settings['items']);

    // Allow modules to alter the order data sent to Affirm before the request.
    $event = new AffirmTransactionDataPreSend($checkout_settings, $order);
    $this->eventDispatcher->dispatch($event, AffirmEvents::AFFIRM_TRANSACTION_DATA_PRESEND);

    $js_settings = [
      'publicKey' => $this->getPublicKey(),
      'locale' => $this->getLocale(),
      'countryCode' => $this->getCountryCode(),
      'scriptUrl' => $this->getScriptUrl(),
      'elementId' => $this->getWindowMode() === 'modal' ? 'edit-actions-next' : 'commerce-checkout-flow',
      'checkoutSettings' => $checkout_settings,
    ];

    if ($this->getLogSettings()) {
      $this->logger->debug('Checkout data for order ID @id: <pre>@data</pre>', [
        '@id' => $order->id(),
        '@data' => print_r($js_settings, TRUE),
      ]);
    }

    return $js_settings;
  }

  /**
   * Turn adjustment data into the corresponding checkout object data.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param \Drupal\commerce_order\Adjustment $adjustment
   *   The adjustment.
   * @param array $data
   *   The Affirm checkout object data as an array.
   */
  protected function handleAffirmCheckoutAdjustments(OrderInterface $order, Adjustment $adjustment, array &$data): void {
    switch ($adjustment->getType()) {
      case 'promotion':
        $promotion = Promotion::load($adjustment->getSourceId());
        if ($promotion->hasCoupons()) {
          /** @var \Drupal\commerce_promotion\Entity\Coupon $coupon */
          foreach ($order->coupons->referencedEntities() as $coupon) {
            if ($coupon->getPromotionId() == $promotion->id()) {
              if (!isset($data['discounts'][$coupon->getCode()])) {
                $data['discounts'][$coupon->getCode()] = [
                  'discount_display_name' => $promotion->getName(),
                  'discount_amount' => '0',
                ];
                break;
              }
            }
          }
          $data['discounts'][$coupon->getCode()]['discount_amount'] = Calculator::add($data['discounts'][$coupon->getCode()]['discount_amount'], Calculator::multiply(-1, $this->minorUnitsConverter->toMinorUnits($adjustment->getAmount())));
        }
        else {
          if (!isset($data['discounts'][$promotion->id()])) {
            $data['discounts'][$promotion->id()] = [
              'discount_display_name' => $promotion->getName(),
              'discount_amount' => '0',
            ];
          }
          $data['discounts'][$promotion->id()]['discount_amount'] = Calculator::add($data['discounts'][$promotion->id()]['discount_amount'], Calculator::multiply(-1, $this->minorUnitsConverter->toMinorUnits($adjustment->getAmount())));
        }
        break;

      case 'shipping':
        $data['shipping_amount'] = Calculator::add($data['shipping_amount'], $this->minorUnitsConverter->toMinorUnits($adjustment->getAmount()));
        break;

      case 'tax':
        $data['tax_amount'] = Calculator::add($data['tax_amount'], $this->minorUnitsConverter->toMinorUnits($adjustment->getAmount()));
        break;
    }
  }

  /**
   * Gets checkout display label.
   *
   * @return array
   *   Checkout display label.
   */
  public function getCheckoutDisplayLabel(): array {
    $prepend_logo = $this->configuration['prepend_display_name_with_logo'];
    if ($prepend_logo == 'none') {
      return [];
    }

    $display_label = '';
    $hide_display_name = $this->configuration['hide_display_name_use_logo'];
    if (!$hide_display_name) {
      $display_label = $this->getDisplayLabel();
    }

    $images_folder = $this->moduleHandler->getModule('commerce_affirm')->getPath() . '/images';
    $svg_content = file_get_contents("$images_folder/$prepend_logo.svg");

    return [
      '#markup' => Markup::create('<span class="affirm-logo-label">' . $svg_content . $display_label . '</span>'),
    ];
  }

}
