<?php

namespace Drupal\commerce_affirm\Plugin\views\area;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\views\Plugin\views\area\AreaPluginBase;
use Drupal\views\Plugin\views\argument\NumericArgument;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines an Affirm area handler for the order total.
 *
 * @ingroup views_area_handlers
 *
 * @ViewsArea("commerce_affirm_monthly_payment_messaging")
 */
class MonthlyPaymentMessaging extends AreaPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The order storage.
   *
   * @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage
   */
  protected $orderStorage;

  /**
   * The product variation storage.
   *
   * @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage
   */
  protected $productVariationStorage;

  /**
   * The minor units converter.
   *
   * @var \Drupal\commerce_price\MinorUnitsConverterInterface
   */
  protected $minorUnitsConverter;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->orderStorage = $container->get('entity_type.manager')->getStorage('commerce_order');
    $instance->productVariationStorage = $container->get('entity_type.manager')->getStorage('commerce_product_variation');
    $instance->minorUnitsConverter = $container->get('commerce_price.minor_units_converter');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions(): array {
    $options = parent::defineOptions();
    $options['page_type'] = [
      'default' => 'cart',
    ];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state): void {
    parent::buildOptionsForm($form, $form_state);

    $form['empty']['#description'] = $this->t("Even if selected, this area handler will never render if a valid order cannot be found in the View's arguments.");
    $form['page_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Page type'),
      '#default_value' => $this->options['page_type'],
      '#options' => commerce_affirm_get_page_types(),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function render($empty = FALSE): array {
    if (!$empty || !empty($this->options['empty'])) {
      $price = NULL;
      $variation = NULL;
      foreach ($this->view->argument as $argument) {
        // First look for an order_id argument.
        if (!$argument instanceof NumericArgument) {
          continue;
        }
        switch ($argument->getField()) {
          case 'commerce_order.order_id':
          case 'commerce_order_item.order_id':
            /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
            $order = $this->orderStorage->load($argument->getValue());
            $price = $order->getTotalPrice();
            break;

          case 'commerce_product_variation.id':
            /** @var \Drupal\commerce_product\Entity\ProductVariationInterface $variation */
            $variation = $this->productVariationStorage->load($argument->getValue());
            $price = $variation->getPrice();
            break;

          default:
            continue 2;
        }
      }
      if ($price) {
        $number = $this->minorUnitsConverter->toMinorUnits($price);
        return [
          '#theme' => 'commerce_affirm_monthly_payment_message',
          '#page_type' => $this->options['page_type'],
          '#number' => $number,
          '#variation' => $variation,
        ];
      }
    }

    return [];
  }

}
